@@ -52,6 +52,8 @@ from The Open Group.
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "libxfontint.h"
+#include "src/util/replace.h"
 
 #include <ctype.h>
 #include <X11/fonts/fntfilst.h>
@@ -106,7 +108,7 @@ bdfReadBitmap(CharInfoPtr pCI, FontFilePtr file, int bit, int byte,
 
     widthBytes = BYTES_PER_ROW(widthBits, glyph);
     if (widthBytes * height > 0) {
-	picture = malloc(widthBytes * height);
+	picture = mallocarray(widthBytes, height);
 	if (!picture) {
           bdfError("Couldn't allocate picture (%d*%d)\n", widthBytes, height);
 	    goto BAILOUT;
@@ -298,7 +300,7 @@ bdfReadCharacters(FontFilePtr file, FontPtr pFont, bdfFileState *pState,
 	bdfError("invalid number of CHARS in BDF file\n");
 	return (FALSE);
     }
-    if (nchars > INT32_MAX / sizeof(CharInfoRec)) {
+    if (nchars > (signed) (INT32_MAX / sizeof(CharInfoRec))) {
 	bdfError("Couldn't allocate pCI (%d*%d)\n", nchars,
 		 (int) sizeof(CharInfoRec));
 	goto BAILOUT;
@@ -312,7 +314,7 @@ bdfReadCharacters(FontFilePtr file, FontPtr pFont, bdfFileState *pState,
     bitmapFont->metrics = ci;
 
     if (bitmapExtra) {
-	bitmapExtra->glyphNames = malloc(nchars * sizeof(Atom));
+	bitmapExtra->glyphNames = mallocarray(nchars, sizeof(Atom));
 	if (!bitmapExtra->glyphNames) {
 	    bdfError("Couldn't allocate glyphNames (%d*%d)\n",
 		     nchars, (int) sizeof(Atom));
@@ -320,7 +322,7 @@ bdfReadCharacters(FontFilePtr file, FontPtr pFont, bdfFileState *pState,
 	}
     }
     if (bitmapExtra) {
-	bitmapExtra->sWidths = malloc(nchars * sizeof(int));
+	bitmapExtra->sWidths = mallocarray(nchars, sizeof(int));
 	if (!bitmapExtra->sWidths) {
 	    bdfError("Couldn't allocate sWidth (%d *%d)\n",
 		     nchars, (int) sizeof(int));
@@ -426,7 +428,7 @@ bdfReadCharacters(FontFilePtr file, FontPtr pFont, bdfFileState *pState,
 	    goto BAILOUT;
 	}
 	/* xCharInfo metrics are stored as INT16 */
-	if ((wx < 0) || (wx > INT16_MAX)) {
+	if ((wx < INT16_MIN) || (wx > INT16_MAX)) {
 	    bdfError("character '%s' has out of range width, %d\n",
 		     charName, wx);
 	    goto BAILOUT;
@@ -631,7 +633,7 @@ bdfReadProperties(FontFilePtr file, FontPtr pFont, bdfFileState *pState)
     }
     if ((sscanf((char *) line, "STARTPROPERTIES %d", &nProps) != 1) ||
 	(nProps <= 0) ||
-	(nProps > ((INT32_MAX / sizeof(FontPropRec)) - BDF_GENPROPS))) {
+	(nProps > (signed) ((INT32_MAX / sizeof(FontPropRec)) - BDF_GENPROPS))) {
 	bdfError("bad 'STARTPROPERTIES'\n");
 	return (FALSE);
     }
@@ -639,7 +641,7 @@ bdfReadProperties(FontFilePtr file, FontPtr pFont, bdfFileState *pState)
     pFont->info.props = NULL;
     pFont->info.nprops = 0;
 
-    stringProps = malloc((nProps + BDF_GENPROPS) * sizeof(char));
+    stringProps = mallocarray((nProps + BDF_GENPROPS), sizeof(char));
     pFont->info.isStringProp = stringProps;
     if (stringProps == NULL) {
 	bdfError("Couldn't allocate stringProps (%d*%d)\n",
