@@ -33,22 +33,8 @@ from The Open Group.
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
-
-/*
- * Translate monolithic #defines to modular definitions
- */
-
-#ifdef PCFFORMAT
-#define XFONT_PCFFORMAT 1
-#endif
-
-#ifdef SNFFORMAT
-#define XFONT_SNFFORMAT 1
-#endif
-
-#ifdef BDFFORMAT
-#define XFONT_BDFFORMAT 1
-#endif
+#include "libxfontint.h"
+#include "src/util/replace.h"
 
 #include <X11/fonts/fntfilst.h>
 #include <X11/fonts/bitmap.h>
@@ -59,15 +45,12 @@ from The Open Group.
 #define   MAX(a,b)    (((a)>(b)) ? a : b)
 #endif
 
-/* Should get this from elsewhere */
-extern unsigned long serverGeneration;
-
 static void bitmapUnloadScalable (FontPtr pFont);
-static void ScaleBitmap ( FontPtr pFont, CharInfoPtr opci, 
-			  CharInfoPtr pci, double *inv_xform, 
+static void ScaleBitmap ( FontPtr pFont, CharInfoPtr opci,
+			  CharInfoPtr pci, double *inv_xform,
 			  double widthMult, double heightMult );
-static FontPtr BitmapScaleBitmaps(FontPtr pf, FontPtr opf, 
-				  double widthMult, double heightMult,   
+static FontPtr BitmapScaleBitmaps(FontPtr pf, FontPtr opf,
+				  double widthMult, double heightMult,
 				  FontScalablePtr vals);
 
 enum scaleType {
@@ -78,17 +61,17 @@ enum scaleType {
 };
 
 typedef struct _fontProp {
-    char       *name;
+    const char *name;
     Atom        atom;
     enum scaleType type;
 } fontProp;
 
-static FontEntryPtr FindBestToScale ( FontPathElementPtr fpe, 
-				      FontEntryPtr entry, 
-				      FontScalablePtr vals, 
-				      FontScalablePtr best, 
-				      double *dxp, double *dyp, 
-				      double *sdxp, double *sdyp, 
+static FontEntryPtr FindBestToScale ( FontPathElementPtr fpe,
+				      FontEntryPtr entry,
+				      FontScalablePtr vals,
+				      FontScalablePtr best,
+				      double *dxp, double *dyp,
+				      double *sdxp, double *sdyp,
 				      FontPathElementPtr *fpep );
 
 static unsigned long bitscaleGeneration = 0;	/* initialization flag */
@@ -231,8 +214,8 @@ get_matrix_vertical_component(double *matrix)
 
 
 static Bool
-ComputeScaleFactors(FontScalablePtr from, FontScalablePtr to, 
-		    double *dx, double *dy, double *sdx, double *sdy, 
+ComputeScaleFactors(FontScalablePtr from, FontScalablePtr to,
+		    double *dx, double *dy, double *sdx, double *sdy,
 		    double *rescale_x)
 {
     double srcpixelset, destpixelset, srcpixel, destpixel;
@@ -318,10 +301,10 @@ else \
     score += (int)(((double)(8 * s)) * m);
 
 static FontEntryPtr
-FindBestToScale(FontPathElementPtr fpe, FontEntryPtr entry, 
-		FontScalablePtr vals, FontScalablePtr best, 
-		double *dxp, double *dyp, 
-		double *sdxp, double *sdyp, 
+FindBestToScale(FontPathElementPtr fpe, FontEntryPtr entry,
+		FontScalablePtr vals, FontScalablePtr best,
+		double *dxp, double *dyp,
+		double *sdxp, double *sdyp,
 		FontPathElementPtr *fpep)
 {
     FontScalableRec temp;
@@ -425,7 +408,7 @@ FindBestToScale(FontPathElementPtr fpe, FontEntryPtr entry,
 		    ((score == best_score) &&
 		     ((dy_amount < best_dy_amount) ||
  		      ((dy_amount == best_dy_amount) &&
- 		       (dx_amount < best_dx_amount))))) 
+ 		       (dx_amount < best_dx_amount)))))
 	    {
 		best_fpe = FontFileBitmapSources.fpe[source];
 	    	best_scaled = scaled;
@@ -525,8 +508,8 @@ doround(double x)
 }
 
 static int
-computeProps(FontPropPtr pf, char *wasStringProp, 
-	     FontPropPtr npf, char *isStringProp, 
+computeProps(FontPropPtr pf, char *wasStringProp,
+	     FontPropPtr npf, char *isStringProp,
 	     unsigned int nprops, double xfactor, double yfactor,
 	     double sXfactor, double sYfactor)
 {
@@ -580,13 +563,13 @@ computeProps(FontPropPtr pf, char *wasStringProp,
 static int
 ComputeScaledProperties(FontInfoPtr sourceFontInfo, /* the font to be scaled */
 			char *name, 		/* name of resulting font */
-			FontScalablePtr vals, 
+			FontScalablePtr vals,
 			double dx, double dy, 	/* scale factors in x and y */
 			double sdx, double sdy, /* directions */
 			long sWidth, 		/* 1000-pixel average width */
-			FontPropPtr *pProps, 	/* returns properties; 
+			FontPropPtr *pProps, 	/* returns properties;
 						   preallocated */
-			char **pIsStringProp) 	/* return booleans; 
+			char **pIsStringProp) 	/* return booleans;
 						   preallocated */
 {
     int         n;
@@ -597,13 +580,13 @@ ComputeScaledProperties(FontInfoPtr sourceFontInfo, /* the font to be scaled */
     char	*isStringProp;
     int		nProps;
 
-    if (bitscaleGeneration != serverGeneration) {
+    if (bitscaleGeneration != __GetServerGeneration()) {
 	initFontPropTable();
-	bitscaleGeneration = serverGeneration;
+	bitscaleGeneration = __GetServerGeneration();
     }
     nProps = NPROPS + 1 + sizeof(fontPropTable) / sizeof(fontProp) +
 			  sizeof(rawFontPropTable) / sizeof(fontProp);
-    fp = malloc(sizeof(FontPropRec) * nProps);
+    fp = mallocarray(sizeof(FontPropRec), nProps);
     *pProps = fp;
     if (!fp) {
 	fprintf(stderr, "Error: Couldn't allocate font properties (%ld*%d)\n",
@@ -614,7 +597,7 @@ ComputeScaledProperties(FontInfoPtr sourceFontInfo, /* the font to be scaled */
     *pIsStringProp = isStringProp;
     if (!isStringProp)
     {
- fprintf(stderr, "Error: Couldn't allocate isStringProp (%d)\n", nProps);
+	fprintf(stderr, "Error: Couldn't allocate isStringProp (%d)\n", nProps);
 	free (fp);
 	return 1;
     }
@@ -633,8 +616,10 @@ ComputeScaledProperties(FontInfoPtr sourceFontInfo, /* the font to be scaled */
 	*isStringProp = 0;
 	switch (fpt->type) {
 	case atom:
-	    fp->value = MakeAtom(ptr1, ptr2 - ptr1, TRUE);
-	    *isStringProp = 1;
+	    if ((ptr1 != NULL) && (ptr2 != NULL)) {
+		fp->value = MakeAtom(ptr1, ptr2 - ptr1, TRUE);
+		*isStringProp = 1;
+	    }
 	    break;
 	case truncate_atom:
 	    for (ptr3 = ptr1; *ptr3; ptr3++)
@@ -692,8 +677,8 @@ ComputeScaledProperties(FontInfoPtr sourceFontInfo, /* the font to be scaled */
 
 
 static int
-compute_xform_matrix(FontScalablePtr vals, double dx, double dy, 
-		     double *xform, double *inv_xform, 
+compute_xform_matrix(FontScalablePtr vals, double dx, double dy,
+		     double *xform, double *inv_xform,
 		     double *xmult, double *ymult)
 {
     double det;
@@ -751,18 +736,20 @@ compute_xform_matrix(FontScalablePtr vals, double dx, double dy,
  *  ScaleFont
  *  returns a pointer to the new scaled font, or NULL (due to AllocError).
  */
+#pragma GCC diagnostic ignored "-Wbad-function-cast"
+
 static FontPtr
-ScaleFont(FontPtr opf,            /* originating font */                    
-	  double widthMult, 	  /* glyphs width scale factor */           
-	  double heightMult, 	  /* glyphs height scale factor */          
-	  double sWidthMult, 	  /* scalable glyphs width scale factor */  
-	  double sHeightMult, 	  /* scalable glyphs height scale factor */ 
-	  FontScalablePtr vals,	                                            
-	  double *newWidthMult,   /* return: X component of glyphs width    
-				     scale factor */                        
-	  double *newHeightMult,  /* return: Y component of glyphs height   
-				     scale factor */                        
-	  long *sWidth)		  /* return: average 1000-pixel width */    
+ScaleFont(FontPtr opf,            /* originating font */
+	  double widthMult, 	  /* glyphs width scale factor */
+	  double heightMult, 	  /* glyphs height scale factor */
+	  double sWidthMult, 	  /* scalable glyphs width scale factor */
+	  double sHeightMult, 	  /* scalable glyphs height scale factor */
+	  FontScalablePtr vals,
+	  double *newWidthMult,   /* return: X component of glyphs width
+				     scale factor */
+	  double *newHeightMult,  /* return: Y component of glyphs height
+				     scale factor */
+	  long *sWidth)		  /* return: average 1000-pixel width */
 {
     FontPtr     pf;
     FontInfoPtr pfi,
@@ -811,8 +798,6 @@ ScaleFont(FontPtr opf,            /* originating font */
        needs to be for the output font */
     if (vals->nranges)
     {
-	int i;
-
 	pfi->allExist = 0;
 	firstCol = 255;
 	lastCol = 0;
@@ -877,7 +862,7 @@ ScaleFont(FontPtr opf,            /* originating font */
     bitmapFont->encoding = 0;
     bitmapFont->bitmapExtra = 0;
     bitmapFont->pDefault = 0;
-    bitmapFont->metrics = malloc(nchars * sizeof(CharInfoRec));
+    bitmapFont->metrics = mallocarray(nchars, sizeof(CharInfoRec));
     if (!bitmapFont->metrics) {
 	fprintf(stderr, "Error: Couldn't allocate metrics (%d*%ld)\n",
 		nchars, (unsigned long)sizeof(CharInfoRec));
@@ -1015,7 +1000,7 @@ ScaleFont(FontPtr opf,            /* originating font */
 		if (pci->metrics.ascent == -pci->metrics.descent)
 		    pci->metrics.ascent++;
 	    }
-    
+
 	    pci++;
 	}
     }
@@ -1039,7 +1024,7 @@ ScaleFont(FontPtr opf,            /* originating font */
 	    	pfi->minbounds.field = pci->metrics.field; \
 	    if (pfi->maxbounds.field < pci->metrics.field) \
 	    	pfi->maxbounds.field = pci->metrics.field
-    
+
 	    MINMAX(leftSideBearing);
 	    MINMAX(rightSideBearing);
 	    MINMAX(ascent);
@@ -1093,7 +1078,7 @@ ScaleFont(FontPtr opf,            /* originating font */
 	    cols = pfi->lastCol - pfi->firstCol + 1;
 	    r = r - pfi->firstRow;
 	    c = c - pfi->firstCol;
-	    bitmapFont->pDefault = 
+	    bitmapFont->pDefault =
                 ACCESSENCODING(bitmapFont->encoding, r * cols + c);
 	}
     }
@@ -1117,7 +1102,7 @@ bail:
 }
 
 static void
-ScaleBitmap(FontPtr pFont, CharInfoPtr opci, CharInfoPtr pci, 
+ScaleBitmap(FontPtr pFont, CharInfoPtr opci, CharInfoPtr pci,
 	    double *inv_xform, double widthMult, double heightMult)
 {
     register char  *bitmap,		/* The bits */
@@ -1192,7 +1177,7 @@ ScaleBitmap(FontPtr pFont, CharInfoPtr opci, CharInfoPtr pci,
 	    /* Looks like we need to anti-alias.  Create a workspace to
 	       contain the grayscale character plus an additional row and
 	       column for scratch */
-	    char_grayscale = malloc((width + 1) * (height + 1));
+	    char_grayscale = mallocarray((width + 1), (height + 1));
 	    if (char_grayscale)
 	    {
 		diffusion_workspace = calloc((newWidth + 2) * 2, sizeof(int));
@@ -1394,7 +1379,7 @@ ScaleBitmap(FontPtr pFont, CharInfoPtr opci, CharInfoPtr pci,
 	    {
 		register int x = xValue >> 16, y = yValue >> 16;
 		int pixvalue, error;
-    
+
 		pixvalue = ((x >= 0 && x < width && y >= 0 && y < height) ?
 			    char_grayscale[x + y * (width + 1)] : 0) +
 			   thisrow[col] / 16;
@@ -1436,7 +1421,7 @@ ScaleBitmap(FontPtr pFont, CharInfoPtr opci, CharInfoPtr pci,
 	    for (col = 0; col < newWidth; col++)
 	    {
 		register int x = xValue >> 16, y = yValue >> 16;
-    
+
 		if (x >= 0 && x < width && y >= 0 && y < height)
 		{
 		    /* Use point-sampling for rescaling. */
@@ -1495,6 +1480,10 @@ BitmapScaleBitmaps(FontPtr pf,          /* scaled font */
     lastRow = pfi->lastRow;
 
     nchars = (lastRow - firstRow + 1) * (lastCol - firstCol + 1);
+    if (nchars <= 0) {
+        goto bail;
+    }
+
     glyph = pf->glyph;
     for (i = 0; i < nchars; i++)
     {
@@ -1543,13 +1532,13 @@ bail:
 
 /* ARGSUSED */
 int
-BitmapOpenScalable (FontPathElementPtr fpe, 
-		    FontPtr *pFont, 
-		    int flags, 
-		    FontEntryPtr entry, 
+BitmapOpenScalable (FontPathElementPtr fpe,
+		    FontPtr *pFont,
+		    int flags,
+		    FontEntryPtr entry,
 		    char *fileName, /* unused */
-		    FontScalablePtr vals, 
-		    fsBitmapFormat format, 
+		    FontScalablePtr vals,
+		    fsBitmapFormat format,
 		    fsBitmapFormatMask fmask,
 		    FontPtr non_cachable_font)	/* We don't do licensing */
 {
@@ -1610,7 +1599,7 @@ BitmapOpenScalable (FontPathElementPtr fpe,
 
     /* Prepare font properties for the new font */
 
-    strcpy (fontName, scaleFrom->name.name);
+    strlcpy (fontName, scaleFrom->name.name, sizeof(fontName));
     FontParseXLFDName (fontName, vals, FONT_XLFD_REPLACE_VALUE);
 
     propCount = ComputeScaledProperties(&sourceFont->info, fontName, vals,
@@ -1620,29 +1609,26 @@ BitmapOpenScalable (FontPathElementPtr fpe,
     if (!sourceFont->refcnt)
 	FontFileCloseFont((FontPathElementPtr) 0, sourceFont);
 
-    if (propCount && (!props || !isStringProp))
-    {
-	font->info.nprops = 0;
-	font->info.props = (FontPropPtr)0;
-	font->info.isStringProp = (char *)0;
-	bitmapUnloadScalable(font);
-	return AllocError;
-    }
-
     font->info.props = props;
     font->info.nprops = propCount;
     font->info.isStringProp = isStringProp;
 
+    if (propCount && (!props || !isStringProp))
+    {
+	bitmapUnloadScalable(font);
+	return AllocError;
+    }
+
     *pFont = font;
     return Successful;
 }
 
 int
-BitmapGetInfoScalable (FontPathElementPtr fpe, 
-		       FontInfoPtr pFontInfo, 
-		       FontEntryPtr entry, 
-		       FontNamePtr fontName, 
-		       char *fileName, 
+BitmapGetInfoScalable (FontPathElementPtr fpe,
+		       FontInfoPtr pFontInfo,
+		       FontEntryPtr entry,
+		       FontNamePtr fontName,
+		       char *fileName,
 		       FontScalablePtr vals)
 {
     FontPtr pfont;
