@@ -53,6 +53,8 @@ in this Software without prior written authorization from The Open Group.
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "libxfontint.h"
+#include "src/util/replace.h"
 
 #ifdef WIN32
 #define _WILLWINSOCK_
@@ -70,6 +72,7 @@ in this Software without prior written authorization from The Open Group.
 #include	"fservestr.h"
 #include	<X11/fonts/fontutil.h>
 #include	<errno.h>
+#include	<limits.h>
 
 #include	<time.h>
 #define Time_t time_t
@@ -91,18 +94,23 @@ in this Software without prior written authorization from The Open Group.
 			     (pci)->descent || \
 			     (pci)->characterWidth)
 
-extern void ErrorF(const char *f, ...);
+/*
+ * SIZEOF(r) is in bytes, length fields in the protocol are in 32-bit words,
+ * so this converts for doing size comparisons.
+ */
+#define LENGTHOF(r)	(SIZEOF(r) >> 2)
+
+/* Somewhat arbitrary limit on maximum reply size we'll try to read. */
+#define MAX_REPLY_LENGTH	((64 * 1024 * 1024) >> 2)
 
 static int fs_read_glyphs ( FontPathElementPtr fpe, FSBlockDataPtr blockrec );
 static int fs_read_list ( FontPathElementPtr fpe, FSBlockDataPtr blockrec );
 static int fs_read_list_info ( FontPathElementPtr fpe,
 			       FSBlockDataPtr blockrec );
 
-extern fd_set _fs_fd_mask;
+static void fs_block_handler ( void *wt );
 
-static void fs_block_handler ( pointer data, OSTimePtr wt,
-			       pointer LastSelectMask );
-static int fs_wakeup ( FontPathElementPtr fpe, unsigned long *mask );
+static int fs_wakeup ( FontPathElementPtr fpe );
 
 /*
  * List of all FPEs
@@ -116,14 +124,14 @@ static CARD32	fs_blockState;
 static int _fs_restart_connection ( FSFpePtr conn );
 static void fs_send_query_bitmaps ( FontPathElementPtr fpe,
 				   FSBlockDataPtr blockrec );
-static int fs_send_close_font ( FontPathElementPtr fpe, Font id );
+static int fs_send_close_font (FSFpePtr conn, Font id);
 static void fs_client_died ( pointer client, FontPathElementPtr fpe );
 static void _fs_client_access ( FSFpePtr conn, pointer client, Bool sync );
 static void _fs_client_resolution ( FSFpePtr conn );
 static fsGenericReply *fs_get_reply (FSFpePtr conn, int *error);
 static int fs_await_reply (FSFpePtr conn);
 static void _fs_do_blocked (FSFpePtr conn);
-static void fs_cleanup_bfont (FSBlockedFontPtr bfont);
+static void fs_cleanup_bfont (FSFpePtr conn, FSBlockedFontPtr bfont);
 
 char _fs_glyph_undefined;
 char _fs_glyph_requested;
@@ -137,7 +145,7 @@ static void
 _fs_close_server (FSFpePtr conn);
 
 static FSFpePtr
-_fs_init_conn (char *servername);
+_fs_init_conn (const char *servername, FontPathElementPtr fpe);
 
 static int
 _fs_wait_connect (FSFpePtr conn);
@@ -157,6 +165,9 @@ _fs_free_conn (FSFpePtr conn);
 static int
 fs_free_fpe(FontPathElementPtr fpe);
 
+static void
+fs_fd_handler(int fd, void *data);
+
 /*
  * Font server access
  *
@@ -206,13 +217,26 @@ _fs_add_rep_log (FSFpePtr conn, fsGenericReply *rep)
 		 rep->sequenceNumber,
 		 conn->reqbuffer[i].opcode);
 }
+
+#define _fs_reply_failed(rep, name, op) do {                            \
+    if (rep) {                                                          \
+        if (rep->type == FS_Error)                                      \
+            fprintf (stderr, "Error: %d Request: %s\n",                 \
+                     ((fsError *)rep)->request, #name);                 \
+        else                                                            \
+            fprintf (stderr, "Bad Length for %s Reply: %u %s %d\n",     \
+                     #name, (unsigned) rep->length, op, LENGTHOF(name));\
+    }                                                                   \
+} while (0)
+
 #else
 #define _fs_add_req_log(conn,op)    ((conn)->current_seq++)
 #define _fs_add_rep_log(conn,rep)
+#define _fs_reply_failed(rep,name,op)
 #endif
 
 static Bool
-fs_name_check(char *name)
+fs_name_check(const char *name)
 {
     /* Just make sure there is a protocol/ prefix */
     return (name && *name != '/' && strchr(name, '/'));
@@ -270,7 +294,7 @@ static int
 fs_init_fpe(FontPathElementPtr fpe)
 {
     FSFpePtr    conn;
-    char       *name;
+    const char  *name;
     int         err;
     int		ret;
 
@@ -282,12 +306,12 @@ fs_init_fpe(FontPathElementPtr fpe)
     if (*name == ':')
 	name++;			/* skip ':' */
 
-    conn = _fs_init_conn (name);
+    conn = _fs_init_conn (name, fpe);
     if (!conn)
 	err = AllocError;
     else
     {
-	err = init_fs_handlers (fpe, fs_block_handler);
+	err = init_fs_handlers2 (fpe, fs_block_handler);
 	if (err != Successful)
 	{
 	    _fs_free_conn (conn);
@@ -360,7 +384,7 @@ fs_free_fpe(FontPathElementPtr fpe)
     }
     _fs_unmark_block (conn, conn->blockState);
     fs_close_conn(conn);
-    remove_fs_handlers(fpe, fs_block_handler, fs_fpes == 0);
+    remove_fs_handlers2(fpe, fs_block_handler, fs_fpes == 0);
     _fs_free_conn (conn);
     fpe->private = (pointer) 0;
 
@@ -492,6 +516,24 @@ _fs_add_clients_depending(FSClientsDependingPtr *clients_depending, pointer clie
     return Suspended;
 }
 
+static void
+conn_start_listening(FSFpePtr conn)
+{
+    if (!conn->fs_listening) {
+	add_fs_fd(conn->fs_fd, fs_fd_handler, conn->fpe);
+	conn->fs_listening = TRUE;
+    }
+}
+
+static void
+conn_stop_listening(FSFpePtr conn)
+{
+    if (conn->fs_listening) {
+	remove_fs_fd(conn->fs_fd);
+	conn->fs_listening = FALSE;
+    }
+}
+
 /*
  * When a request is aborted due to a font server failure,
  * signal any depending clients to restart their dependant
@@ -504,7 +546,7 @@ _fs_clean_aborted_blockrec(FSFpePtr conn, FSBlockDataPtr blockrec)
     case FS_OPEN_FONT: {
 	FSBlockedFontPtr bfont = (FSBlockedFontPtr)blockrec->data;
 
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
 	_fs_signal_clients_depending(&bfont->clients_depending);
 	break;
     }
@@ -523,7 +565,7 @@ _fs_clean_aborted_blockrec(FSFpePtr conn, FSBlockDataPtr blockrec)
 	FSBlockedListInfoPtr binfo;
 	binfo = (FSBlockedListInfoPtr) blockrec->data;
 	if (binfo->status == FS_LFWI_REPLY)
-	    FD_SET(conn->fs_fd, &_fs_fd_mask);
+	    conn_start_listening(conn);
 	_fs_free_props (&binfo->info);
     }
     default:
@@ -543,16 +585,12 @@ fs_abort_blockrec(FSFpePtr conn, FSBlockDataPtr blockrec)
  * then unload the partially created font
  */
 static void
-fs_cleanup_bfont (FSBlockedFontPtr bfont)
+fs_cleanup_bfont (FSFpePtr conn, FSBlockedFontPtr bfont)
 {
-    FSFontDataRec *fsd;
-
     if (bfont->pfont)
     {
-	fsd = (FSFontDataRec *) bfont->pfont->fpePrivate;
-
 	/* make sure the FS knows we choked on it */
-	fs_send_close_font(bfont->pfont->fpe, bfont->fontid);
+	fs_send_close_font(conn, bfont->fontid);
 
 	/*
 	 * Either unload the font if it's being opened for
@@ -570,7 +608,10 @@ fs_cleanup_bfont (FSBlockedFontPtr bfont)
 	    bfont->pfont = 0;
 	}
 	else
+	{
+	    FSFontDataRec *fsd = (FSFontDataRec *)bfont->pfont->fpePrivate;
 	    fsd->generation = -1;
+	}
     }
 }
 
@@ -585,7 +626,7 @@ fs_get_reply (FSFpePtr conn, int *error)
     int		    ret;
 
     /* block if the connection is down or paused in lfwi */
-    if (conn->fs_fd == -1 || !FD_ISSET (conn->fs_fd, &_fs_fd_mask))
+    if (conn->fs_fd == -1 || !conn->fs_listening)
     {
 	*error = FSIO_BLOCK;
 	return 0;
@@ -600,6 +641,21 @@ fs_get_reply (FSFpePtr conn, int *error)
 
     rep = (fsGenericReply *) buf;
 
+    /*
+     * Refuse to accept replies longer than a maximum reasonable length,
+     * before we pass to _fs_start_read, since it will try to resize the
+     * incoming connection buffer to this size.  Also avoids integer overflow
+     * on 32-bit systems.
+     */
+    if (rep->length > MAX_REPLY_LENGTH)
+    {
+	ErrorF("fserve: reply length %ld > MAX_REPLY_LENGTH, disconnecting"
+	       " from font server\n", (long)rep->length);
+	_fs_connection_died (conn);
+	*error = FSIO_ERROR;
+	return 0;
+    }
+
     ret = _fs_start_read (conn, rep->length << 2, &buf);
     if (ret != FSIO_READY)
     {
@@ -617,7 +673,7 @@ fs_reply_ready (FSFpePtr conn)
 {
     fsGenericReply  *rep;
 
-    if (conn->fs_fd == -1 || !FD_ISSET (conn->fs_fd, &_fs_fd_mask))
+    if (conn->fs_fd == -1 || !conn->fs_listening)
 	return FALSE;
     if (fs_data_read (conn) < sizeof (fsGenericReply))
 	return FALSE;
@@ -682,13 +738,15 @@ fs_read_open_font(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     int			    ret;
 
     rep = (fsOpenBitmapFontReply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	(rep->length != LENGTHOF(fsOpenBitmapFontReply)))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	if (rep)
 	    _fs_done_read (conn, rep->length << 2);
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
+	_fs_reply_failed (rep, fsOpenBitmapFontReply, "!=");
 	return BadFontName;
     }
 
@@ -703,7 +761,7 @@ fs_read_open_font(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 
     if (rep->otherid && !(bfont->flags & FontReopen))
     {
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
 
 	/* Find old font if we're completely done getting it from server. */
 	bfont->pfont = find_old_font(rep->otherid);
@@ -815,6 +873,7 @@ fs_read_query_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     FSFpePtr		conn = (FSFpePtr) fpe->private;
     fsQueryXInfoReply	*rep;
     char		*buf;
+    long		bufleft; /* length of reply left to use */
     fsPropInfo		*pi;
     fsPropOffset	*po;
     pointer		pd;
@@ -824,13 +883,15 @@ fs_read_query_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     int			ret;
 
     rep = (fsQueryXInfoReply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	(rep->length < LENGTHOF(fsQueryXInfoReply)))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	if (rep)
 	    _fs_done_read (conn, rep->length << 2);
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
+	_fs_reply_failed (rep, fsQueryXInfoReply, "<");
 	return BadFontName;
     }
 
@@ -844,6 +905,9 @@ fs_read_query_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     buf = (char *) rep;
     buf += SIZEOF(fsQueryXInfoReply);
 
+    bufleft = rep->length << 2;
+    bufleft -= SIZEOF(fsQueryXInfoReply);
+
     /* move the data over */
     fsUnpack_XFontInfoHeader(rep, pInfo);
 
@@ -851,22 +915,55 @@ fs_read_query_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     _fs_init_fontinfo(conn, pInfo);
 
     /* Compute offsets into the reply */
+    if (bufleft < SIZEOF(fsPropInfo))
+    {
+	ret = -1;
+#ifdef DEBUG
+	fprintf(stderr, "fsQueryXInfo: bufleft (%ld) < SIZEOF(fsPropInfo)\n",
+		bufleft);
+#endif
+	goto bail;
+    }
     pi = (fsPropInfo *) buf;
     buf += SIZEOF (fsPropInfo);
+    bufleft -= SIZEOF(fsPropInfo);
 
+    if ((bufleft / SIZEOF(fsPropOffset)) < pi->num_offsets)
+    {
+	ret = -1;
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXInfo: bufleft (%ld) / SIZEOF(fsPropOffset) < %u\n",
+		bufleft, (unsigned) pi->num_offsets);
+#endif
+	goto bail;
+    }
     po = (fsPropOffset *) buf;
     buf += pi->num_offsets * SIZEOF(fsPropOffset);
+    bufleft -= pi->num_offsets * SIZEOF(fsPropOffset);
 
+    if (bufleft < pi->data_len)
+    {
+	ret = -1;
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXInfo: bufleft (%ld) < data_len (%u)\n",
+		bufleft, (unsigned) pi->data_len);
+#endif
+	goto bail;
+    }
     pd = (pointer) buf;
     buf += pi->data_len;
+    bufleft -= pi->data_len;
 
     /* convert the properties and step over the reply */
     ret = _fs_convert_props(pi, po, pd, pInfo);
+  bail:
     _fs_done_read (conn, rep->length << 2);
 
     if (ret == -1)
     {
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
 	return AllocError;
     }
 
@@ -885,7 +982,7 @@ fs_read_query_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 	}
 	else
 	{
-	    fs_cleanup_bfont (bfont);
+	    fs_cleanup_bfont (conn, bfont);
 	    err = BadFontName;
 	}
 	_fs_free_props (pInfo);
@@ -951,13 +1048,15 @@ fs_read_extent_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     FontInfoRec		    *fi = &bfont->pfont->info;
 
     rep = (fsQueryXExtents16Reply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	(rep->length < LENGTHOF(fsQueryXExtents16Reply)))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	if (rep)
 	    _fs_done_read (conn, rep->length << 2);
-	fs_cleanup_bfont (bfont);
+	fs_cleanup_bfont (conn, bfont);
+	_fs_reply_failed (rep, fsQueryXExtents16Reply, "<");
 	return BadFontName;
     }
 
@@ -970,12 +1069,31 @@ fs_read_extent_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 	numInfos *= 2;
 	haveInk = TRUE;
     }
-    ci = pCI = malloc(sizeof(CharInfoRec) * numInfos);
+    if (numInfos >= (INT_MAX / sizeof(CharInfoRec))) {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXExtents16: numInfos (%d) >= %ld\n",
+		numInfos, (INT_MAX / sizeof(CharInfoRec)));
+#endif
+	pCI = NULL;
+    }
+    else if (numExtents > ((rep->length - LENGTHOF(fsQueryXExtents16Reply))
+			    / LENGTHOF(fsXCharInfo))) {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXExtents16: numExtents (%d) > (%u - %d) / %d\n",
+		numExtents, (unsigned) rep->length,
+		LENGTHOF(fsQueryXExtents16Reply), LENGTHOF(fsXCharInfo));
+#endif
+	pCI = NULL;
+    }
+    else
+	pCI = mallocarray(numInfos, sizeof(CharInfoRec));
 
     if (!pCI)
     {
 	_fs_done_read (conn, rep->length << 2);
-	fs_cleanup_bfont(bfont);
+	fs_cleanup_bfont(conn, bfont);
 	return AllocError;
     }
     fsfont->encoding = pCI;
@@ -1111,7 +1229,7 @@ fs_read_extent_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 }
 
 #ifdef DEBUG
-static char *fs_open_states[] = {
+static const char *fs_open_states[] = {
     "OPEN_REPLY  ",
     "INFO_REPLY  ",
     "EXTENT_REPLY",
@@ -1200,15 +1318,12 @@ _fs_unmark_block (FSFpePtr conn, CARD32 mask)
 
 /* ARGSUSED */
 static void
-fs_block_handler(pointer data, OSTimePtr wt, pointer LastSelectMask)
+fs_block_handler(void *wt)
 {
-    static struct timeval block_timeout;
     CARD32	now, earliest, wakeup;
     int		soonest;
     FSFpePtr    conn;
 
-    XFD_ORSET((fd_set *)LastSelectMask, (fd_set *)LastSelectMask,
-	      &_fs_fd_mask);
     /*
      * Flush all pending output
      */
@@ -1220,14 +1335,7 @@ fs_block_handler(pointer data, OSTimePtr wt, pointer LastSelectMask)
      * Check for any fpe with a complete reply, set sleep time to zero
      */
     if (fs_blockState & FS_COMPLETE_REPLY)
-    {
-	block_timeout.tv_sec = 0;
-	block_timeout.tv_usec = 0;
-	if (*wt == NULL)
-	    *wt = &block_timeout;
-	else
-	    **wt = block_timeout;
-    }
+	adjust_fs_wait_for_delay(wt, 0);
     /*
      * Walk through fpe list computing sleep time
      */
@@ -1268,12 +1376,7 @@ fs_block_handler(pointer data, OSTimePtr wt, pointer LastSelectMask)
 	soonest = earliest - now;
 	if (soonest < 0)
 	    soonest = 0;
-	block_timeout.tv_sec = soonest / 1000;
-	block_timeout.tv_usec = (soonest % 1000) * 1000;
-	if (*wt == NULL)
-	    *wt = &block_timeout;
-	else if (soonest < (*wt)->tv_sec * 1000 + (*wt)->tv_usec / 1000)
-	    **wt = block_timeout;
+	adjust_fs_wait_for_delay(wt, soonest);
     }
 }
 
@@ -1360,11 +1463,11 @@ fs_read_reply (FontPathElementPtr fpe, pointer client)
     }
 }
 
-static int
-fs_wakeup(FontPathElementPtr fpe, unsigned long *mask)
+static void
+fs_fd_handler(int fd, void *data)
 {
-    fd_set	    *LastSelectMask = (fd_set *) mask;
-    FSFpePtr	    conn = (FSFpePtr) fpe->private;
+    FontPathElementPtr	fpe = data;
+    FSFpePtr	    	conn = (FSFpePtr) fpe->private;
 
     /*
      * Don't continue if the fd is -1 (which will be true when the
@@ -1372,16 +1475,23 @@ fs_wakeup(FontPathElementPtr fpe, unsigned long *mask)
      */
     if ((conn->blockState & FS_RECONNECTING))
 	_fs_check_reconnect (conn);
-    else if ((conn->blockState & FS_COMPLETE_REPLY) ||
-	     (conn->fs_fd != -1 && FD_ISSET(conn->fs_fd, LastSelectMask)))
+    else if ((conn->fs_fd != -1))
 	fs_read_reply (fpe, 0);
+}
+
+static int
+fs_wakeup(FontPathElementPtr fpe)
+{
+    FSFpePtr	    conn = (FSFpePtr) fpe->private;
+
     if (conn->blockState & (FS_PENDING_REPLY|FS_BROKEN_CONNECTION|FS_BROKEN_WRITE))
 	_fs_do_blocked (conn);
+    if (conn->blockState & FS_COMPLETE_REPLY)
+	fs_read_reply (fpe, 0);
 #ifdef DEBUG
     {
 	FSBlockDataPtr	    blockrec;
 	FSBlockedFontPtr    bfont;
-	FSBlockedListPtr    blist;
 	static CARD32	    lastState;
 	static FSBlockDataPtr	lastBlock;
 
@@ -1405,7 +1515,6 @@ fs_wakeup(FontPathElementPtr fpe, unsigned long *mask)
 			 "<freed>");
 		break;
 	    case FS_LIST_FONTS:
-		blist = (FSBlockedListPtr) blockrec->data;
 		fprintf (stderr, "  Blocked list errcode %d sequence %d\n",
 			 blockrec->errcode, blockrec->sequenceNumber);
 		break;
@@ -1517,7 +1626,7 @@ _fs_do_blocked (FSFpePtr conn)
 /* ARGSUSED */
 static int
 fs_send_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
-		  char *name, int namelen,
+		  const char *name, int namelen,
 		  fsBitmapFormat format, fsBitmapFormatMask fmask,
 		  XID id, FontPtr *ppfont)
 {
@@ -1535,7 +1644,7 @@ fs_send_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
     if (conn->blockState & FS_GIVE_UP)
 	return BadFontName;
 
-    if (namelen <= 0 || namelen > sizeof (buf) - 1)
+    if (namelen < 0 || namelen > sizeof (buf) - 1)
 	return BadFontName;
 
     /*
@@ -1587,16 +1696,6 @@ fs_send_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
 	return AllocError;
     }
 
-    /*
-     * Must check this before generating any protocol, otherwise we'll
-     * mess up a reconnect in progress
-     */
-    if (conn->blockState & (FS_BROKEN_CONNECTION | FS_RECONNECTING))
-    {
-	_fs_pending_reply (conn);
-	return Suspended;
-    }
-
     fsd->generation = conn->generation;
 
     bfont = (FSBlockedFontPtr) blockrec->data;
@@ -1608,6 +1707,16 @@ fs_send_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
     bfont->clients_depending = (FSClientsDependingPtr)0;
     bfont->freeFont = (flags & FontReopen) == 0;
 
+    /*
+     * Must check this before generating any protocol, otherwise we'll
+     * mess up a reconnect in progress
+     */
+    if (conn->blockState & (FS_BROKEN_CONNECTION | FS_RECONNECTING))
+    {
+	_fs_pending_reply (conn);
+	return Suspended;
+    }
+
     _fs_client_access (conn, client, (flags & FontOpenSync) != 0);
     _fs_client_resolution(conn);
 
@@ -1678,7 +1787,7 @@ fs_send_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
 	if (err == Successful)
 	    *ppfont = bfont->pfont;
 	else
-	    fs_cleanup_bfont (bfont);
+	    fs_cleanup_bfont (conn, bfont);
 	bfont->freeFont = FALSE;
 	_fs_remove_block_rec (conn, blockrec);
     }
@@ -1709,7 +1818,7 @@ fs_send_query_bitmaps(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 /* ARGSUSED */
 static int
 fs_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
-	     char *name, int namelen,
+	     const char *name, int namelen,
 	     fsBitmapFormat format, fsBitmapFormatMask fmask,
 	     XID id, FontPtr *ppfont,
 	     char **alias, FontPtr non_cachable_font)
@@ -1735,7 +1844,7 @@ fs_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
 	    if (err == Successful)
 		*ppfont = bfont->pfont;
 	    else
-		fs_cleanup_bfont (bfont);
+		fs_cleanup_bfont (conn, bfont);
 	    _fs_remove_block_rec (conn, blockrec);
 	    return err;
 	}
@@ -1746,9 +1855,8 @@ fs_open_font(pointer client, FontPathElementPtr fpe, Mask flags,
 
 /* ARGSUSED */
 static int
-fs_send_close_font(FontPathElementPtr fpe, Font id)
+fs_send_close_font(FSFpePtr conn, Font id)
 {
-    FSFpePtr    conn = (FSFpePtr) fpe->private;
     fsCloseReq  req;
 
     if (conn->blockState & FS_GIVE_UP)
@@ -1772,7 +1880,7 @@ fs_close_font(FontPathElementPtr fpe, FontPtr pfont)
     FSFpePtr	    conn = (FSFpePtr) fpe->private;
 
     if (conn->generation == fsd->generation)
-	fs_send_close_font(fpe, fsd->fontid);
+	fs_send_close_font(conn, fsd->fontid);
 
 #ifdef DEBUG
     {
@@ -1809,6 +1917,7 @@ fs_read_glyphs(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     FontInfoPtr		    pfi = &pfont->info;
     fsQueryXBitmaps16Reply  *rep;
     char		    *buf;
+    long		    bufleft; /* length of reply left to use */
     fsOffset32		    *ppbits;
     fsOffset32		    local_off;
     char		    *off_adr;
@@ -1825,22 +1934,48 @@ fs_read_glyphs(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     unsigned long	    minchar, maxchar;
 
     rep = (fsQueryXBitmaps16Reply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	(rep->length < LENGTHOF(fsQueryXBitmaps16Reply)))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	if (rep)
 	    _fs_done_read (conn, rep->length << 2);
 	err = AllocError;
+	_fs_reply_failed (rep, fsQueryXBitmaps16Reply, "<");
 	goto bail;
     }
 
     buf = (char *) rep;
     buf += SIZEOF (fsQueryXBitmaps16Reply);
 
+    bufleft = rep->length << 2;
+    bufleft -= SIZEOF (fsQueryXBitmaps16Reply);
+
+    if ((bufleft / SIZEOF (fsOffset32)) < rep->num_chars)
+    {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXBitmaps16: num_chars (%u) > bufleft (%ld) / %d\n",
+		(unsigned) rep->num_chars, bufleft, SIZEOF (fsOffset32));
+#endif
+	err = AllocError;
+	goto bail;
+    }
     ppbits = (fsOffset32 *) buf;
     buf += SIZEOF (fsOffset32) * (rep->num_chars);
+    bufleft -= SIZEOF (fsOffset32) * (rep->num_chars);
 
+    if (bufleft < rep->nbytes)
+    {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsQueryXBitmaps16: nbytes (%u) > bufleft (%ld)\n",
+		(unsigned) rep->nbytes, bufleft);
+#endif
+	err = AllocError;
+	goto bail;
+    }
     pbitmaps = (pointer ) buf;
 
     if (blockrec->type == FS_LOAD_GLYPHS)
@@ -1898,7 +2033,9 @@ fs_read_glyphs(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 	     */
 	    if (NONZEROMETRICS(&fsdata->encoding[minchar].metrics))
 	    {
-		if (local_off.length)
+		if (local_off.length &&
+		    (local_off.position < rep->nbytes) &&
+		    (local_off.length <= (rep->nbytes - local_off.position)))
 		{
 		    bits = allbits;
 		    allbits += local_off.length;
@@ -1940,7 +2077,8 @@ fs_read_glyphs(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     err = Successful;
 
 bail:
-    _fs_done_read (conn, rep->length << 2);
+    if (rep)
+	_fs_done_read (conn, rep->length << 2);
     return err;
 }
 
@@ -2033,11 +2171,6 @@ fs_send_load_glyphs(pointer client, FontPtr pfont,
     return Suspended;
 }
 
-
-extern pointer serverClient;	/* This could be any number that
-				   doesn't conflict with existing
-				   client values. */
-
 static int
 _fs_load_glyphs(pointer client, FontPtr pfont, Bool range_flag,
 		unsigned int nchars, int item_size, unsigned char *data)
@@ -2206,17 +2339,17 @@ fs_load_all_glyphs(FontPtr pfont)
      * perform an unpleasant job that, we hope, will never be required.
      */
 
-    while ((err = _fs_load_glyphs(serverClient, pfont, TRUE, 0, 0, NULL)) ==
+    while ((err = _fs_load_glyphs(__GetServerClient(), pfont, TRUE, 0, 0, NULL)) ==
 	   Suspended)
     {
 	if (fs_await_reply (conn) != FSIO_READY)
 	{
 	    /* Get rid of blockrec */
-	    fs_client_died(serverClient, pfont->fpe);
+	    fs_client_died(__GetServerClient(), pfont->fpe);
 	    err = BadCharRange;
 	    break;
 	}
-	fs_read_reply (pfont->fpe, serverClient);
+	fs_read_reply (pfont->fpe, __GetServerClient());
     }
     return err;
 }
@@ -2228,38 +2361,55 @@ fs_read_list(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     FSBlockedListPtr	blist = (FSBlockedListPtr) blockrec->data;
     fsListFontsReply	*rep;
     char		*data;
+    long		dataleft; /* length of reply left to use */
     int			length,
 			i,
 			ret;
     int			err;
 
     rep = (fsListFontsReply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	(rep->length < LENGTHOF(fsListFontsReply)))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	if (rep)
 	    _fs_done_read (conn, rep->length << 2);
+	_fs_reply_failed (rep, fsListFontsReply, "<");
 	return AllocError;
     }
     data = (char *) rep + SIZEOF (fsListFontsReply);
+    dataleft = (rep->length << 2) - SIZEOF (fsListFontsReply);
 
     err = Successful;
     /* copy data into FontPathRecord */
     for (i = 0; i < rep->nFonts; i++)
     {
+	if (dataleft < 1)
+	    break;
 	length = *(unsigned char *)data++;
-	err = AddFontNamesName(blist->names, data, length);
+	dataleft--; /* used length byte */
+	if (length > dataleft) {
+#ifdef DEBUG
+	    fprintf(stderr,
+		    "fsListFonts: name length (%d) > dataleft (%ld)\n",
+		    length, dataleft);
+#endif
+	    err = BadFontName;
+	    break;
+	}
+	err = xfont2_add_font_names_name(blist->names, data, length);
 	if (err != Successful)
 	    break;
 	data += length;
+	dataleft -= length;
     }
     _fs_done_read (conn, rep->length << 2);
     return err;
 }
 
 static int
-fs_send_list_fonts(pointer client, FontPathElementPtr fpe, char *pattern,
+fs_send_list_fonts(pointer client, FontPathElementPtr fpe, const char *pattern,
 		   int patlen, int maxnames, FontNamesPtr newnames)
 {
     FSFpePtr		conn = (FSFpePtr) fpe->private;
@@ -2315,7 +2465,7 @@ fs_send_list_fonts(pointer client, FontPathElementPtr fpe, char *pattern,
 
 static int
 fs_list_fonts(pointer client, FontPathElementPtr fpe,
-	      char *pattern, int patlen, int maxnames, FontNamesPtr newnames)
+	      const char *pattern, int patlen, int maxnames, FontNamesPtr newnames)
 {
     FSFpePtr		conn = (FSFpePtr) fpe->private;
     FSBlockDataPtr	blockrec;
@@ -2347,6 +2497,7 @@ fs_read_list_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     FSBlockedListInfoPtr	binfo = (FSBlockedListInfoPtr) blockrec->data;
     fsListFontsWithXInfoReply	*rep;
     char			*buf;
+    long			bufleft;
     FSFpePtr			conn = (FSFpePtr) fpe->private;
     fsPropInfo			*pi;
     fsPropOffset		*po;
@@ -2358,12 +2509,15 @@ fs_read_list_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     _fs_free_props (&binfo->info);
 
     rep = (fsListFontsWithXInfoReply *) fs_get_reply (conn, &ret);
-    if (!rep || rep->type == FS_Error)
+    if (!rep || rep->type == FS_Error ||
+	((rep->nameLength != 0) &&
+	 (rep->length < LENGTHOF(fsListFontsWithXInfoReply))))
     {
 	if (ret == FSIO_BLOCK)
 	    return StillWorking;
 	binfo->status = FS_LFWI_FINISHED;
 	err = AllocError;
+	_fs_reply_failed (rep, fsListFontsWithXInfoReply, "<");
 	goto done;
     }
     /*
@@ -2380,6 +2534,7 @@ fs_read_list_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
     }
 
     buf = (char *) rep + SIZEOF (fsListFontsWithXInfoReply);
+    bufleft = (rep->length << 2) - SIZEOF (fsListFontsWithXInfoReply);
 
     /*
      * The original FS implementation didn't match
@@ -2388,19 +2543,71 @@ fs_read_list_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
      */
     if (conn->fsMajorVersion <= 1)
     {
+	if (rep->nameLength > bufleft) {
+#ifdef DEBUG
+	    fprintf(stderr,
+		    "fsListFontsWithXInfo: name length (%d) > bufleft (%ld)\n",
+		    (int) rep->nameLength, bufleft);
+#endif
+	    err = AllocError;
+	    goto done;
+	}
+	/* binfo->name is a 256 char array, rep->nameLength is a CARD8 */
 	memcpy (binfo->name, buf, rep->nameLength);
 	buf += _fs_pad_length (rep->nameLength);
+	bufleft -= _fs_pad_length (rep->nameLength);
     }
     pi = (fsPropInfo *) buf;
+    if (SIZEOF (fsPropInfo) > bufleft) {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsListFontsWithXInfo: PropInfo length (%d) > bufleft (%ld)\n",
+		(int) SIZEOF (fsPropInfo), bufleft);
+#endif
+	err = AllocError;
+	goto done;
+    }
+    bufleft -= SIZEOF (fsPropInfo);
     buf += SIZEOF (fsPropInfo);
     po = (fsPropOffset *) buf;
+    if (pi->num_offsets > (bufleft / SIZEOF (fsPropOffset))) {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsListFontsWithXInfo: offset length (%u * %d) > bufleft (%ld)\n",
+		(unsigned) pi->num_offsets, (int) SIZEOF (fsPropOffset), bufleft);
+#endif
+	err = AllocError;
+	goto done;
+    }
+    bufleft -= pi->num_offsets * SIZEOF (fsPropOffset);
     buf += pi->num_offsets * SIZEOF (fsPropOffset);
     pd = (pointer) buf;
+    if (pi->data_len > bufleft) {
+#ifdef DEBUG
+	fprintf(stderr,
+		"fsListFontsWithXInfo: data length (%u) > bufleft (%ld)\n",
+		(unsigned) pi->data_len, bufleft);
+#endif
+	err = AllocError;
+	goto done;
+    }
+    bufleft -= pi->data_len;
     buf += pi->data_len;
     if (conn->fsMajorVersion > 1)
     {
+	if (rep->nameLength > bufleft) {
+#ifdef DEBUG
+	    fprintf(stderr,
+		    "fsListFontsWithXInfo: name length (%d) > bufleft (%ld)\n",
+		    (int) rep->nameLength, bufleft);
+#endif
+	    err = AllocError;
+	    goto done;
+	}
+	/* binfo->name is a 256 char array, rep->nameLength is a CARD8 */
 	memcpy (binfo->name, buf, rep->nameLength);
 	buf += _fs_pad_length (rep->nameLength);
+	bufleft -= _fs_pad_length (rep->nameLength);
     }
 
 #ifdef DEBUG
@@ -2420,7 +2627,7 @@ fs_read_list_info(FontPathElementPtr fpe, FSBlockDataPtr blockrec)
 
     /* disable this font server until we've processed this response */
     _fs_unmark_block (conn, FS_COMPLETE_REPLY);
-    FD_CLR(conn->fs_fd, &_fs_fd_mask);
+    conn_stop_listening(conn);
 done:
     _fs_done_read (conn, rep->length << 2);
     return err;
@@ -2429,7 +2636,7 @@ done:
 /* ARGSUSED */
 static int
 fs_start_list_with_info(pointer client, FontPathElementPtr fpe,
-			char *pattern, int len, int maxnames, pointer *pdata)
+			const char *pattern, int len, int maxnames, pointer *pdata)
 {
     FSFpePtr		    conn = (FSFpePtr) fpe->private;
     FSBlockDataPtr	    blockrec;
@@ -2523,7 +2730,7 @@ fs_next_list_with_info(pointer client, FontPathElementPtr fpe,
     *numFonts = binfo->remaining;
 
     /* Restart reply processing from this font server */
-    FD_SET(conn->fs_fd, &_fs_fd_mask);
+    conn_start_listening(conn);
     if (fs_reply_ready (conn))
 	_fs_mark_block (conn, FS_COMPLETE_REPLY);
 
@@ -2649,14 +2856,12 @@ _fs_client_access (FSFpePtr conn, pointer client, Bool sync)
 	if (crac.num_auths == 0) {
 	    authorizations = padding;
 	    authlen = 4;
-	} else {
-	    authlen = (authlen + 3) & ~0x3;
 	}
 	crac.length = (sizeof (fsCreateACReq) + authlen) >> 2;
 	crac.acid = cur->acid;
 	_fs_add_req_log(conn, FS_CreateAC);
 	_fs_write(conn, (char *) &crac, sizeof (fsCreateACReq));
-	_fs_write(conn, authorizations, authlen);
+	_fs_write_pad(conn, authorizations, authlen);
 	/* ignore reply; we don't even care about it */
 	conn->curacid = 0;
 	cur->auth_generation = client_auth_generation(client);
@@ -2686,7 +2891,7 @@ _fs_check_connect (FSFpePtr conn)
     switch (ret) {
     case FSIO_READY:
 	conn->fs_fd = _FontTransGetConnectionNumber (conn->trans_conn);
-	FD_SET (conn->fs_fd, &_fs_fd_mask);
+	conn_start_listening(conn);
 	break;
     case FSIO_BLOCK:
 	break;
@@ -2786,7 +2991,7 @@ _fs_recv_conn_setup (FSFpePtr conn)
     int			ret = FSIO_ERROR;
     fsConnSetup		*setup;
     FSFpeAltPtr		alts;
-    int			i, alt_len;
+    unsigned int	i, alt_len;
     int			setup_len;
     char		*alt_save, *alt_names;
 
@@ -2813,8 +3018,9 @@ _fs_recv_conn_setup (FSFpePtr conn)
 	}
 	if (setup->num_alternates)
 	{
+	    size_t alt_name_len = setup->alternate_len << 2;
 	    alts = malloc (setup->num_alternates * sizeof (FSFpeAltRec) +
-			   (setup->alternate_len << 2));
+			   alt_name_len);
 	    if (alts)
 	    {
 		alt_names = (char *) (setup + 1);
@@ -2823,10 +3029,25 @@ _fs_recv_conn_setup (FSFpePtr conn)
 		{
 		    alts[i].subset = alt_names[0];
 		    alt_len = alt_names[1];
+		    if (alt_len >= alt_name_len) {
+			/*
+			 * Length is longer than setup->alternate_len
+			 * told us to allocate room for, assume entire
+			 * alternate list is corrupted.
+			 */
+#ifdef DEBUG
+			fprintf (stderr,
+				 "invalid alt list (length %lx >= %lx)\n",
+				 (long) alt_len, (long) alt_name_len);
+#endif
+			free(alts);
+			return FSIO_ERROR;
+		    }
 		    alts[i].name = alt_save;
 		    memcpy (alt_save, alt_names + 2, alt_len);
 		    alt_save[alt_len] = '\0';
 		    alt_save += alt_len + 1;
+		    alt_name_len -= alt_len + 1;
 		    alt_names += _fs_pad_length (alt_len + 2);
 		}
 		conn->numAlts = setup->num_alternates;
@@ -2964,6 +3185,7 @@ _fs_send_cat_sync (FSFpePtr conn)
      * by a bogus catalogue
      */
     lcreq.reqType = FS_ListCatalogues;
+    lcreq.data = 0;
     lcreq.length = (SIZEOF(fsListCataloguesReq)) >> 2;
     lcreq.maxNames = 0;
     lcreq.nbytes = 0;
@@ -3010,7 +3232,7 @@ _fs_close_server (FSFpePtr conn)
     }
     if (conn->fs_fd >= 0)
     {
-	FD_CLR (conn->fs_fd, &_fs_fd_mask);
+	conn_stop_listening(conn);
 	conn->fs_fd = -1;
     }
     conn->fs_conn_state = FS_CONN_UNCONNECTED;
@@ -3144,11 +3366,12 @@ _fs_start_reconnect (FSFpePtr conn)
 
 
 static FSFpePtr
-_fs_init_conn (char *servername)
+_fs_init_conn (const char *servername, FontPathElementPtr fpe)
 {
     FSFpePtr	conn;
+    size_t	snlen = strlen (servername) + 1;
 
-    conn = calloc (1, sizeof (FSFpeRec) + strlen (servername) + 1);
+    conn = calloc (1, sizeof (FSFpeRec) + snlen);
     if (!conn)
 	return 0;
     if (!_fs_io_init (conn))
@@ -3159,7 +3382,8 @@ _fs_init_conn (char *servername)
     conn->servername = (char *) (conn + 1);
     conn->fs_conn_state = FS_CONN_UNCONNECTED;
     conn->fs_fd = -1;
-    strcpy (conn->servername, servername);
+    conn->fpe = fpe;
+    strlcpy (conn->servername, servername, snlen);
     return conn;
 }
 
@@ -3177,22 +3401,27 @@ _fs_free_conn (FSFpePtr conn)
  * called at server init time
  */
 
+static const xfont2_fpe_funcs_rec fs_fpe_funcs = {
+	.version = XFONT2_FPE_FUNCS_VERSION,
+	.name_check = fs_name_check,
+	.init_fpe = fs_init_fpe,
+	.free_fpe = fs_free_fpe,
+	.reset_fpe = fs_reset_fpe,
+	.open_font = fs_open_font,
+	.close_font = fs_close_font,
+	.list_fonts = fs_list_fonts,
+	.start_list_fonts_with_info = fs_start_list_with_info,
+	.list_next_font_with_info = fs_next_list_with_info,
+	.wakeup_fpe = fs_wakeup,
+	.client_died = fs_client_died,
+	.load_glyphs = _fs_load_glyphs,
+	.start_list_fonts_and_aliases = (StartLaFunc) 0,
+	.list_next_font_or_alias = (NextLaFunc) 0,
+	.set_path_hook = (SetPathFunc) 0
+};
+
 void
 fs_register_fpe_functions(void)
 {
-    RegisterFPEFunctions(fs_name_check,
-			 fs_init_fpe,
-			 fs_free_fpe,
-			 fs_reset_fpe,
-			 fs_open_font,
-			 fs_close_font,
-			 fs_list_fonts,
-			 fs_start_list_with_info,
-			 fs_next_list_with_info,
-			 fs_wakeup,
-			 fs_client_died,
-			 _fs_load_glyphs,
-			 NULL,
-			 NULL,
-			 NULL);
+    register_fpe_funcs(&fs_fpe_funcs);
 }
