@@ -31,8 +31,10 @@ in this Software without prior written authorization from The Open Group.
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "libxfontint.h"
 #include    <X11/fonts/fntfilst.h>
 #include    <X11/keysym.h>
+#include "src/util/replace.h"
 
 #if HAVE_STDINT_H
 #include <stdint.h>
@@ -47,7 +49,7 @@ FontFileInitTable (FontTablePtr table, int size)
 	return FALSE;
     if (size)
     {
-	table->entries = malloc(sizeof(FontEntryRec) * size);
+	table->entries = mallocarray(size, sizeof(FontEntryRec));
 	if (!table->entries)
 	    return FALSE;
     }
@@ -123,10 +125,7 @@ FontFileMakeDir(const char *dirName, int size)
 	dirlen = strlen(dirName);
 	attriblen = 0;
     }
-    if (dirName[dirlen - 1] != '/')
-#ifdef NCD
-    if (dirlen)     /* leave out slash for builtins */
-#endif
+    if (dirlen && dirName[dirlen - 1] != '/')
 	needslash = 1;
     dir = malloc(sizeof *dir + dirlen + needslash + 1 +
 		 (attriblen ? attriblen + 1 : 0));
@@ -151,11 +150,11 @@ FontFileMakeDir(const char *dirName, int size)
     else
 	dir->attributes = NULL;
     strncpy(dir->directory, dirName, dirlen);
-    dir->directory[dirlen] = '\0';
-    if (dir->attributes)
-	strcpy(dir->attributes, attrib);
     if (needslash)
-	strcat(dir->directory, "/");
+	dir->directory[dirlen] = '/';
+    dir->directory[dirlen + needslash] = '\0';
+    if (dir->attributes)
+	strlcpy(dir->attributes, attrib, attriblen + 1);
     return dir;
 }
 
@@ -177,8 +176,13 @@ FontFileAddEntry(FontTablePtr table, FontEntryPtr prototype)
     if (table->sorted)
 	return (FontEntryPtr) 0;    /* "cannot" happen */
     if (table->used == table->size) {
+	if (table->size >= ((INT32_MAX / sizeof(FontEntryRec)) - 100))
+	    /* If we've read so many entries we're going to ask for 2gb
+	       or more of memory, something is so wrong with this font
+	       directory that we should just give up before we overflow. */
+	    return NULL;
 	newsize = table->size + 100;
-	entry = realloc(table->entries, newsize * sizeof(FontEntryRec));
+	entry = reallocarray(table->entries, newsize, sizeof(FontEntryRec));
 	if (!entry)
 	    return (FontEntryPtr)0;
 	table->size = newsize;
@@ -394,8 +398,10 @@ PatternMatch(char *pat, int patdashes, char *string, int stringdashes)
 		}
 	    }
 	case '?':
-	    if (*string++ == XK_minus)
+	    if ((t = *string++) == XK_minus)
 		stringdashes--;
+	    if (!t)
+		return 0;
 	    break;
 	case '\0':
 	    return (*string == '\0');
@@ -508,7 +514,7 @@ FontFileFindNamesInScalableDir(FontTablePtr table, FontNamePtr pat, int max,
 	{
 	    name = &table->entries[i].name;
 	    if (newmax) *newmax = max - 1;
-	    return AddFontNamesName(names, name->name, name->length);
+	    return xfont2_add_font_names_name(names, name->name, name->length);
 	}
 	start = i;
 	stop = i + 1;
@@ -545,7 +551,7 @@ FontFileFindNamesInScalableDir(FontTablePtr table, FontNamePtr pat, int max,
 		    continue;
 	    }
 
-	    ret = AddFontNamesName(names, fname->name.name, fname->name.length);
+	    ret = xfont2_add_font_names_name(names, fname->name.name, fname->name.length);
 	    if (ret != Successful)
 		goto bail;
 
@@ -557,7 +563,7 @@ FontFileFindNamesInScalableDir(FontTablePtr table, FontNamePtr pat, int max,
 	    {
 		names->length[names->nnames - 1] =
 		    -names->length[names->nnames - 1];
-		ret = AddFontNamesName(names, fname->u.alias.resolved,
+		ret = xfont2_add_font_names_name(names, fname->u.alias.resolved,
 				       strlen(fname->u.alias.resolved));
 		if (ret != Successful)
 		    goto bail;
